# coding: utf-8

import click
import os
import peewee
import re

from ..models import Album
from ..models import Picture
from ..models import ResizedPicture
from ..models import Size

from ..utils import fatal_error


MOVE_MODELS = [
    (Album, 'album_path'),
    (Picture, 'file_path'),
    (ResizedPicture, 'picture_path'),
]


class Clean(object):

    albums_removed = 0
    pictures_checked = 0
    resized_deleted = 0
    keep_resized = False
    resize_dir = None
    page = 0

    def __init__(self, config):
        self.config = config

    def _remove_picture(self, pic):
        nr = pic.resized_pictures.count()
        pic.albums.clear()
        pic.delete_instance(recursive=not self.keep_resized)
        if nr and not self.keep_resized:
            self.resized_deleted += 1

    def _get_model_instances(self, model, page=0):
        page += 1
        return model.select().order_by(model.id).paginate(page, 1000), page

    def _clean_model_by_path(self, model, path_name, removal_method=None):
        checked = 0
        removed = 0
        instances, page = self._get_model_instances(model)
        while instances.count():
            for instance in instances:
                checked += 1
                path = getattr(instance, path_name)
                if not os.path.exists(path):
                    # click.echo("%s: %s" % (model, path))
                    # continue
                    if removal_method is None:
                        instance.delete_instance()
                    else:
                        self.__getattribute__(removal_method)(instance)
                    removed += 1
            instances, page = self._get_model_instances(model, page)
        return checked, removed

    def _delete_resizes(self):
        checked, deleted = 0, 0
        available_sizes = [size.name for size in Size.select()]
        for path, dirs, files in os.walk(self.resize_dir):
            rel_path = os.path.relpath(path, self.resize_dir)
            for f in files:
                checked += 1
                m = re.match(
                    '^(?P<key>[a-zA-Z0-9]{6})-(?P<size_name>\w+)\.{1}[a-z]+$',
                    f
                )
                if m:
                    if m.group('size_name') not in available_sizes:
                        click.echo("Size does not exist: %s" % os.path.join(path, f))
                        # os.unlink(os.path.join(path, f))
                        deleted += 1
                        continue
                    key = '%s%s' % (rel_path, m.group('key'))
                    try:
                        Picture.get(Picture.key == key)
                    except peewee.DoesNotExist:
                        click.echo("Key (%s) does not exist: %s" % (key, os.path.join(path, f)))
                        # os.unlink(os.path.join(path, f))
                        deleted += 1
        return checked, deleted

    def clean(self, keep_resized=False, resize_dir=None):
        self.resize_dir = self.config.resized_photos_dir if \
            resize_dir is None else resize_dir
        self.keep_resized = keep_resized
        # Cleaning pictures
        click.echo("Cleaning pictures... ", nl=False)
        checked, removed = self._clean_model_by_path(
            Picture, 'file_path', '_remove_picture'
        )
        click.echo("Done.")
        click.echo("    Removed %s of the %s pictures." % (removed, checked))
        if self.resized_deleted:
            click.echo("    Deleted %s resized versions." %
                       self.resized_deleted)

        # Removing deleted resizes
        click.echo("Cleaning resizes deleted on the system... ", nl=False)
        checked, removed = self._clean_model_by_path(
            ResizedPicture, 'path',
        )
        click.echo("Done.")
        click.echo("    Removed %s of the %s resizes." % (removed, checked))

        # Removing orphaned resizes
        if self.resize_dir is not None:
            click.echo("Deleting orphaned resized pictures...", nl=False)
            checked, deleted = self._delete_resizes()
            click.echo("Done.")
            click.echo("    Deleted %s orphaned resizes." % deleted)

        # Removing deleted albums
        click.echo("Cleaning folder based albums... ", nl=False)
        checked, removed = self._clean_model_by_path(Album, 'album_path')
        click.echo("Done.")
        click.echo("    Removed %s of the %s albums." % (removed, checked))

    def move(self, from_dir=None, to_dir=None):
        click.echo('Changing all save paths that start with:')
        click.echo('    %s' % from_dir)
        click.echo('to:')
        click.echo('    %s' % to_dir)
        if from_dir is None or to_dir is None:
            return fatal_error('Please provide both an old directory and a '
                               'new directory.')
        from_dir = from_dir[:-1] if from_dir[-1] == '/' else from_dir
        to_dir = to_dir[:-1] if to_dir[-1] == '/' else to_dir
        counter = 0
        for model, path in MOVE_MODELS:
            instances = model.select().where(
                getattr(model, path).startswith(from_dir)
            )
            for instance in instances:
                old_path = getattr(instance, path)
                setattr(instance, path, old_path.replace(from_dir, to_dir))
                instance.save()
                counter += 1
        click.echo('Done.')
        click.echo('Replaced %s instances.' % counter)



import random
import numpy

from PIL import Image


def most_frequent_colour(image):

    w, h = image.size
    pixels = image.getcolors(w * h)

    most_frequent_pixel = pixels[0]

    for count, colour in pixels:
        if count > most_frequent_pixel[0]:
            most_frequent_pixel = (count, colour)

    return most_frequent_pixel


def average_colour(image):
    colour_tuple = [None, None, None]
    for channel in range(3):

        # Get data for one channel at a time
        pixels = image.getdata(band=channel)

        values = []
        for pixel in pixels:
            values.append(pixel)

        colour_tuple[channel] = sum(values) / len(values)

    return tuple(colour_tuple)


class Cluster(object):

    def __init__(self):
        self.pixels = []
        self.centroid = None

    def add_point(self, pixel):
        self.pixels.append(pixel)

    def set_new_centroid(self):

        r = [colour[0] for colour in self.pixels]
        g = [colour[1] for colour in self.pixels]
        b = [colour[2] for colour in self.pixels]

        r = sum(r) / len(r)
        g = sum(g) / len(g)
        b = sum(b) / len(b)

        self.centroid = (r, g, b)
        self.pixels = []

        return self.centroid


class Kmeans(object):

    def __init__(self, k=3, max_iterations=5, min_distance=5.0, size=200):
        self.k = k
        self.max_iterations = max_iterations
        self.min_distance = min_distance
        self.size = (size, size)

    def run(self, image):
        self.image = image
        self.image.thumbnail(self.size)
        self.pixels = numpy.array(image.getdata(), dtype=numpy.uint8)

        self.clusters = [None for i in range(self.k)]
        self.old_clusters = None

        random_pixels = random.sample(self.pixels, self.k)

        for idx in range(self.k):
            self.clusters[idx] = Cluster()
            self.clusters[idx].centroid = random_pixels[idx]

        iterations = 0

        while self.should_exit(iterations) is False:

            self.old_clusters = [cluster.centroid for cluster in self.clusters]

            print(iterations)

            for pixel in self.pixels:
                self.assign_clusters(pixel)

            for cluster in self.clusters:
                cluster.set_new_centroid()

            iterations += 1

        return [cluster.centroid for cluster in self.clusters]

    def assign_clusters(self, pixel):
        shortest = float('Inf')
        for cluster in self.clusters:
            distance = self.calc_distance(cluster.centroid, pixel)
            if distance < shortest:
                shortest = distance
                nearest = cluster

        nearest.addPoint(pixel)

    def calc_distance(self, a, b):

        result = numpy.sqrt(sum((a - b) ** 2))
        return result

    def should_exit(self, iterations):

        if self.old_clusters is None:
            return False

        for idx in range(self.k):
            dist = self.calc_distance(
                numpy.array(self.clusters[idx].centroid),
                numpy.array(self.old_clusters[idx])
            )
            if dist < self.min_distance:
                return True

        if iterations <= self.max_iterations:
            return False

        return True

    # ############################################
    # The remaining methods are used for debugging
    def show_image(self):
        self.image.show()

    def show_centroid_colours(self):

        for cluster in self.clusters:
            image = Image.new("RGB", (200, 200), cluster.centroid)
            image.show()

    def show_clustering(self):

        local_pixels = [None] * len(self.image.getdata())

        for idx, pixel in enumerate(self.pixels):
                shortest = float('Inf')
                for cluster in self.clusters:
                    distance = self.calc_distance(
                        cluster.centroid,
                        pixel
                    )
                    if distance < shortest:
                        shortest = distance
                        nearest = cluster

                local_pixels[idx] = nearest.centroid

        w, h = self.image.size
        local_pixels = numpy.asarray(local_pixels)\
            .astype('uint8')\
            .reshape((h, w, 3))

        colour_map = Image.fromarray(local_pixels)
        colour_map.show()



# coding: utf-8

import math
from decimal import Decimal

try:
    from PIL import Image
except ImportError:
    import Image


def resize_and_cache(org_file, new_file, new_width, new_height,
                     crop=False, rotation=0):
    """ Creates a resized image of the original file and saves it
        to a new location
        If crop is true the image will be cropped to fit the exact sizes
        otherwise the image will fit within the sizes.
    """
    try:
        im = Image.open(org_file)
    except IOError:
        return False, 'Could not open orignal file.'

    if rotation:
        im = im.rotate(rotation)

    arorg = Decimal(im.size[0]) / Decimal(im.size[1])
    arnew = (Decimal(new_width) / Decimal(new_height))

    if crop:
        # The orginal image is wider
        if arorg > arnew:
            width = im.size[1] * arnew
            left = int(math.fabs((im.size[0] - width) / 2))
            box = (left, 0, im.size[0] - left, im.size[1])
        # The original image is taller
        elif arorg < arnew:
            height = im.size[0] / arnew
            top = int(math.fabs((im.size[1] - height) / 2))
            box = (0, top, im.size[0], im.size[1] - top)
        # The original picture has the same aspect ration
        else:
            box = (0, 0, im.size[0], im.size[1])
        new = im.crop(box)
        new.thumbnail((new_width, new_height), Image.ANTIALIAS)

    else:
        # check if the image is larger or not then the image we want to
        # resize to.
        if im.size[0] > new_height or im.size[1] > new_height:
            # default
            width = Decimal(new_width)
            height = Decimal(new_height)
            # The orginal image is wider
            if arorg > arnew:
                height = Decimal(new_height) / arorg
            # The original image is taller
            elif arorg < arnew:
                width = Decimal(new_width) / arorg
            new = im.crop((0, 0, im.size[0], im.size[1]))
            new.thumbnail((width, height), Image.ANTIALIAS)
        else:
            new = im

    try:
        new.save(new_file, "JPEG", quality=95)
    except IOError:
        return False, 'Could not save resized image file.'

    width, height = new.size[0], new.size[1]
    del(im, new)
    return (width, height), False



# coding: utf-8

import click
import datetime
import peewee
import pickle
import yaml

from ..classes import Config
from ..models import Album
from ..utils import alphanum_hash
from ..utils import fatal_error
from ..utils.aliased_group import AliasedGroup


def _date_from_input(date_str):
    try:
        return datetime.datetime.strptime("%s" % date_str, "%Y-%m-%d")
    except ValueError:
        try:
            return datetime.datetime.strptime("%s" % date_str, "%Y-%m-%d-%H-%M")
        except ValueError:
            return False
    return None


def _validate_date(ctx, param, value):
    if value is None:
        return None
    date = _date_from_input(value)
    if date is False:
        now = datetime.datetime.now()
        raise click.BadParameter(
            'Please use YYYY-MM-DD or YYYY-MM-DD-hh-mm as a date format. '
            'E.g. %s or %s for now.' % (
                now.strftime('%Y-%m-%d'),
                now.strftime('%Y-%m-%d-%H-%M')
            )
        )
    else:
        return date


@click.group(
    'album',
    cls=AliasedGroup,
    help="Album management",
    epilog="Use \"htg album COMMAND --help\" for help with subcommands.",
)
def album(*args, **kwargs):
    pass


@click.command('list', help='List Albums.')
@click.option(
    '--year', '-y',
    help='Only list albums of a certain year.',
    type=int,
    # callback=_test_year,
)
@click.option(
    '--title', '-t',
    help='Filter title by string.',
    # callback=_test_year,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def list(config_file=None, year=None, title=None, *args, **kwargs):
    Config(config_file)
    albums = Album.select()
    if year is not None:
        albums = albums.where(
            Album.start_date > datetime.datetime(year, 1, 1),
            Album.start_date < datetime.datetime(year + 1, 1, 1),
        )
    if title is not None:
        albums = albums.where(Album.title.contains(title))
    if not albums.count():
        click.echo("No albums found.")
        return
    click.echo('┌' + (40 * '─') + '┬' + (12 * '─'), nl=False)
    click.echo('┬' + (7 * '─') + '┬' + (12 * '─') + '┬───┐')
    click.echo('│ {:<38} │ '.format('TITLE'), nl=False)
    click.echo('{:^10} │ '.format('START DATE'), nl=False)
    click.echo('{:<5} │ '.format('PICS'), nl=False)
    click.echo('{:<10} │ * │'.format('KEY'))
    click.echo('├' + (40 * '─') + '┼' + (12 * '─'), nl=False)
    click.echo('┼' + (7 * '─') + '┼' + (12 * '─') + '┼───┤')
    for a in albums:
        title = a.title if len(a.title) < 39 else "%s..." % a.title[:35]
        start_date = a.start_date.strftime("%Y-%m-%d") if a.start_date else '-'
        nr = a.pictures.count() if a.from_dir else '?'
        click.echo('│ ', nl=False)
        click.echo('{:<38}'.format(title), nl=False)
        click.echo(' │ ', nl=False)
        click.echo('{:^10}'.format(start_date), nl=False)
        click.echo(' │ ', nl=False)
        click.echo('{:>5}'.format(nr), nl=False)
        click.echo(' │ ', nl=False)
        click.echo('{:<10}'.format(a.key), nl=False)
        click.echo(' │ ', nl=False)
        click.echo('-' if a.from_dir else '✓', nl=False)
        click.echo(' │')
    click.echo('└' + (40 * '─') + '┴' + (12 * '─'), nl=False)
    click.echo('┴' + (7 * '─') + '┴' + (12 * '─') + '┴───┘')
    click.echo('* Editable? Only albums based on date ranges can be edited.')
    click.echo('  Other albums should be edited through the meta file.')

album.add_command(list)


@click.command('show', help='Show all info for an Album.')
@click.argument(
    'key',
    required=True,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def show(key, config_file=None, *args, **kwargs):
    Config(config_file)
    try:
        a = Album.get(key=key)
    except peewee.DoesNotExist:
        return fatal_error('No album with key: %s' % key)
    if a.start_date and a.end_date:
        date = "%s to %s" % (
            a.start_date.strftime('%b %d, %Y'),
            a.end_date.strftime('%b %d, %Y')
        )
    elif a.start_date:
        date = a.start_date.strftime('%b %d, %Y')
    else:
        date = 'no date'
    click.echo('Key: %s' % a.key)
    click.echo('Title: %s' % a.title)
    click.echo('Date: %s' % date)
    click.echo('Nr of photos: %s' % a.pictures.count())
    click.echo('Album directory: %s' % a.album_path)
    meta = pickle.loads(a.meta.encode()) if a.meta else None
    if meta:
        click.echo("Other meta data below line.")
        click.echo(80 * "─")
        click.echo(yaml.dump(meta, default_flow_style=False))

album.add_command(show)


@click.command('add', help='Add an Album based on a date range.')
@click.option(
    '--title', '-t',
    required=True,
    help='Title of your album',
)
@click.option(
    '--end-date', '-e',
    required=False,
    help='End date (and time) for photos in this album '
         '(YYYY-MM-DD or YYYY-MM-DD-hh-mm)',
    callback=_validate_date,
)
@click.option(
    '--start-date', '-s',
    required=True,
    help='Start date (and time) for photos in this album '
         '(YYYY-MM-DD or YYYY-MM-DD-hh-mm)',
    callback=_validate_date,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def add(config_file=None, *args, **kwargs):
    Config(config_file)
    a = Album.create(
        from_dir=False,
        album_path=alphanum_hash(length=32),
        start_date=kwargs['start_date'],
        end_date=kwargs.get('end_date', None),
        title=kwargs['title'],
    )
    click.echo('Album was saved with key: %s' % a.key)

album.add_command(add)


@click.command('edit', help='Edit an Album based on a date range.')
@click.option(
    '--title', '-t',
    required=False,
    help='New title for the selected album',
)
@click.option(
    '--end-date', '-e',
    required=False,
    help='New end date (and time) for photos in the selected album '
         '(YYYY-MM-DD or YYYY-MM-DD-hh-mm)',
    callback=_validate_date,
)
@click.option(
    '--start-date', '-s',
    required=False,
    help='New start date (and time) for photos in the selected album '
         '(YYYY-MM-DD or YYYY-MM-DD-hh-mm)',
    callback=_validate_date,
)
@click.argument(
    'key',
    required=True,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def edit(key, config_file=None, *args, **kwargs):
    Config(config_file)
    try:
        a = Album.get(key=key, from_dir=False)
    except peewee.DoesNotExist:
        return fatal_error('No editable album with key: %s' % key)
    click.echo('%s' % kwargs)
    if kwargs.get('title', None):
        a.title = kwargs['title']
    if kwargs.get('end_date', None):
        a.end_date = kwargs['end_date']
    if kwargs.get('start_date', None):
        click.echo('Updating start_date')
        a.start_date = kwargs['start_date']
    a.save()
    click.echo("Album was updated.")

album.add_command(edit)


@click.command(
    'delete',
    help='Delete an Album. Note that albums based on a meta file in or '
         'directory will be added again after you scan the directory.'
)
@click.argument(
    'key',
    required=True,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def delete(key, config_file=None, *args, **kwargs):
    Config(config_file)
    try:
        a = Album.get(key=key)
    except peewee.DoesNotExist:
        return fatal_error('No album with key: %s' % key)
    if click.confirm(
        'Are you sure you want to delete the album with title: \n%s' % a.title
    ):
        a.delete_instance(recursive=True)
        click.echo("Album deleted.")
    else:
        click.echo("Aborted.")

album.add_command(delete)



import click

from ..classes import Config
from ..migrate.in2ex import in2ex as in2exfunc
from ..utils.aliased_group import AliasedGroup


@click.group(
    'migrate',
    cls=AliasedGroup,
    help="Migration functions to help with upgrades",
    epilog="Use \"htg utils migrate COMMAND --help\" for help with "
           "subcommands.",
)
def migrate(*args, **kwargs):
    pass


@click.command('in2ex', help="Move from include to exclude in meta_file.")
@click.option(
    '--directory', '-d',
    required=False,
    metavar="DIR",
    multiple=True,
    help='The in2ex will default to the original_photos_dirs setting '
         'in the config file if you do not provide root directories.',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def in2ex(config_file=None, directory=[], include_all=False):
    config = Config(config_file)
    dirs = directory if directory else config.original_photos_dirs
    in2exfunc(config, dirs)

migrate.add_command(in2ex)



# -*- coding: utf8 -*-

import click
import os

from ..models import Picture
from ..utils import abort
from ..utils import check_and_create_dir
from ..imaging import resize_and_cache


class ResizeSimple(object):

    queue = []
    sizes = {}
    resizes_processed = 0
    page = 0

    def __init__(self, config, force=False, sizes=[], output_dir=None):
        self.config = config
        self.output_dir = output_dir if output_dir else \
            self.config.resized_photos_dir
        self.sizes = [self.config.sizes[s] for s in self.config.sizes if
                      s in sizes or not sizes]

    def start(self):
        aborted = False
        click.echo("Filling up resize queue... ", nl=False)
        try:
            self._scan_pictures()
        except KeyboardInterrupt:
            return abort("\nCancelled queueing. Nothing changed.")
        click.echo("Done.")

        click.echo("Resizing %s pictures." % len(self.queue))
        try:
            with click.progressbar(self.queue) as bar:
                for job in bar:
                    self._process_resize_job(job)
                    self.resizes_processed += 1
        except KeyboardInterrupt:
            aborted = True
        if aborted:
            msg = "PLEASE NOTE: Aborted after processing %s of %s resizes.\n" \
                  % (self.resizes_processed, len(self.queue))
        else:
            msg = "Done resizing pictures: "
            msg += "Processed %s resizes." % self.resizes_processed
        click.echo(msg)

    def _get_pictures(self):
        self.page += 1
        return Picture.select().order_by(Picture.id).paginate(self.page, 1000)

    def _scan_pictures(self):
        pics = self._get_pictures()
        while pics.count():
            for pic in pics:
                for size in self.sizes:
                    self.queue.append((pic, size))
            pics = self._get_pictures()

    def _resize_path(self, picture, size):
        path, ext = os.path.splitext(picture.file_path)
        return os.path.join(
            self.output_dir,
            picture.key[:2],
            "%s-%s%s" % (
                picture.key[2:],
                size.name,
                ext,
            ),
        )

    def _process_resize_job(self, job):
        picture, size = job
        resize_path = self._resize_path(picture, size)
        if os.path.isfile(resize_path):
            return True
        resize_dir, name = os.path.split(resize_path)
        if check_and_create_dir(resize_dir, create=True):
            success, msg = resize_and_cache(
                self.config.resize_with,
                picture.file_path,
                resize_path,
                size.width,
                size.height,
                crop=size.crop,
            )
            return success
        return False



import datetime
import os

import peewee as pw

from playhouse.shortcuts import model_to_dict

from .utils import alphanum_hash
from .utils import slugify
from .utils import to_utc_timestamp

database = pw.SqliteDatabase(None)


class BaseModel(pw.Model):
    def dict(self):
        return model_to_dict(self)

    class Meta:
        database = database


class Picture(BaseModel):
    key = pw.CharField(unique=True, default=alphanum_hash)
    # Initial fields, must always be set.
    checksum = pw.CharField(unique=True)
    file_path = pw.TextField(unique=True)
    file_name = pw.CharField()

    # Meta data fields, only set on change
    sort_date = pw.DateTimeField(null=True)
    caption = pw.CharField(null=True)
    latitude = pw.FloatField(null=True)
    longitude = pw.FloatField(null=True)
    rotation = pw.IntegerField(default=True)

    # last_scan = pw.ForeignKeyField(Scan, related_name='pictures', null=True)
    last_change_indexed = pw.DateTimeField(default=datetime.datetime.now)

    def build_dict(self, picture_root):
        d = {
            'name': self.file_name,
            'ext': self.ext,
            'path': self.resize_path(picture_root),
        }
        if self.sort_date is not None:
            d['date'] = to_utc_timestamp(self.sort_date)
        if self.caption:
            d['caption'] = self.caption
        if self.latitude is not None and self.longitude is not None:
            d['location'] = [self.latitude, self.longitude]
        return d

    @property
    def ext(self):
        name, ext = os.path.splitext(self.file_name)
        return ext.lower()

    def resize_path(self, picture_root):
        return os.path.join(picture_root, self.key[:2], "%s-" % self.key[2:],)

    class Meta:
        db_table = 'pictures'


class Size(BaseModel):
    name = pw.CharField(unique=True)
    width = pw.IntegerField()
    height = pw.IntegerField()
    crop = pw.BooleanField(default=True)
    include_in_json = pw.BooleanField(default=False)
    change_date = pw.DateTimeField(default=datetime.datetime.now)

    def __str__(self):
        return "<Size: %s>" % self.build_dict()

    def build_dict(self):
        return {
            'name': self.name,
            'width': self.width,
            'height': self.height,
            'cropped': self.crop,
            'include_in_json': self.include_in_json,
            'change_date': self.change_date,
        }

    @staticmethod
    def all_names():
        try:
            return [size.name for size in Size.select()]
        except pw.OperationalError:
            return []

    @staticmethod
    def get_size_by_name(name):
        try:
            return Size.get(name=name)
        except Size.DoesNotExist:
            return None

    class Meta:
        db_table = 'sizes'


class ResizedPicture(BaseModel):
    picture = pw.ForeignKeyField(Picture, related_name="resized_pictures")
    size = pw.ForeignKeyField(Size, related_name="sizes")
    picture_path = pw.TextField()
    resize_date = pw.DateTimeField(null=True)
    successfully_created = pw.BooleanField(default=False)
    width = pw.IntegerField(default=0)
    height = pw.IntegerField(default=0)

    @property
    def path(self):
        name, ext = os.path.splitext(self.picture.file_name)
        return os.path.join(
            self.picture_path,
            self.picture.key[:2],
            "%s-%s%s" % (
                self.picture.key[2:],
                self.size.name,
                ext.lower(),
            ),
        )

    def url(self, picture_root):
        name, ext = os.path.splitext(self.picture.file_name)
        return os.path.join(
            picture_root,
            self.picture.key[:2],
            "%s-%s%s" % (
                self.picture.key[2:],
                self.size.name,
                ext,
            ),
        )

    class Meta:
        db_table = 'resized_pictures'


class Album(BaseModel):
    key = pw.CharField(unique=True, default=alphanum_hash)
    hidden = pw.BooleanField(default=False)
    from_dir = pw.BooleanField(default=True)
    album_path = pw.TextField(unique=True)
    start_date = pw.DateTimeField(null=True)
    end_date = pw.DateTimeField(null=True)
    title = pw.CharField()
    description = pw.TextField(null=True)
    meta = pw.TextField(null=True)
    pictures = pw.ManyToManyField(Picture, backref="albums")

    exclude = []

    @property
    def slug(self):
        s = "%s-%s" % (slugify(self.title), self.key)
        if self.start_date:
            s = "%s-%s" % (self.start_date.strftime("%Y-%m-%d"), s)
        return s

    class Meta:
        db_table = 'albums'

AlbumPicture = Album.pictures.get_through_model()


def init_database(db_file):
    database.init(db_file)
    database.connect()
    for tcl in [
        Picture, Size, ResizedPicture, Album, AlbumPicture
    ]:
        if not tcl.table_exists():
            tcl.create_table()
    return database



# coding: utf-8

DEFAULT_CONFIG = """
###############################################################################
# Compulsory settings
###############################################################################

# The directory where htg will save resized photos
# e.g. /resized/photos/go/in/here
resized_photos_dir:

# The directory where htg will save the json data files
# e.g. /json/files/go/in/here
json_data_dir:

# The (relative) url where the webserver can find the resized photos
# e.g. /photos/
resized_photos_url: /photos/

# The (relative) url where the webserver can find the json data files
# e.g. /data/
json_data_url: /data/


###############################################################################
# Optional settings
###############################################################################

# The location of your database file.
# e.g.: /path/where/you/want/to/save/database.sqlite
# If you do not specify the database file the database will be saved as
# db.sqlite in the same directory as the config file.
database_file:

# A list of directories that contain your original photos.
# e.g. /path/where/you/keep/your/awesome/originals
original_photos_dirs:
    -

# Specify the sizes for the images you want to resize
# Make up a name for each size (e.g. large, thumb, etc)
# width: The maximum width of the resized image.
# height: The maximum height of the resized image. Leave empty to use the width.
# crop: Crop to fit maximum dimensions (true, false)
# include_resized_size_in_json: If true the exact size of the resized photo
# will be included in the json output.
sizes:
  large:
    crop: false
    height:
    include_resized_size_in_json: false
    width: 1920
  medium:
    crop: false
    height:
    include_resized_size_in_json: false
    width: 1280
  small:
    crop: false
    height:
    include_resized_size_in_json: false
    width: 640
  thumb:
    crop: false
    height: 150
    include_resized_size_in_json: true
    width: 50000

# By default the picture date is not set if no date can be found in a.
# the exif data and b. the filename. Use this flag to set the date to
# the file creation date.
use_file_creation_date_as_image_date: false

# Look for a file name <meta_file_name> in each directory. This meta file
# can contain settings and album data for that directory. See the
# documentation for info on the meta file.
meta_file_name: _meta.yaml

# Albums can be split into paragraphs and chapters defined inside the meta
# file, but if you set this to true paragraphs will automatically be created
# for each day inside an album. This helps make large amounts of pictures
# easier to digest
# You can also set this value inside the metafile for an album.
auto_day_paragraphs: true

# Albums are created for directories that contain a meta file but can also be
# created based on the directory name that resembles this one:
# YYYY-MM-DD - Title of you Album
create_albums_from_dir_names: true

# For the timeline view you can choose to show all photos or just those photos
# that do not belong to an album.
album_photos_in_timeline: true

# Files that match the following patterns will not be indexed. Files starting
# with . or : are always excluded.
exclude_files_and_dirs_that_match:
  - ^_.+$   # files starting with _

# Files and subdirectories in a directory that contains a file matching the
# name below will not be indexed.
exclude_file_name: _exclude

"""

# All the following settings can be set in the yaml file too.
# However they should be used carefully and do not need to be in the default
# yaml file that is generated from the above.
DEFAULT_ADVANCED = """
###############################################################################
# Other, more advanced settings
###############################################################################
image_extensions:
  - .jpg
  - .jpeg
  - .png
  - .bmp

video_extensions:
  - .mpg
  - .mpeg
  - .mov
  - .m4v
  - .mp4
  - .avi

file_date_formats:
  - \%Y-\%m-\%d-\%H.\%M.\%S
  - PANO_\%Y\%m\%d_\%H\%M\%S
  - TRIM_\%Y\%m\%d_\%H\%M\%S
  - VID_\%Y\%m\%d_\%H\%M\%S
  - IMG_\%Y\%m\%d_\%H\%M\%S
  - \%Y-\%m-\%d \%H.\%M.\%S
  - \%Y\%m\%d\%H\%M\%S
  - \%Y\%m\%d-\%H\%M\%S
  - \%Y\%m\%d-\%H\%M
  - \%Y\%m\%d\%H\%M\%S

album_dir_pattern: ^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})[ ]*-[ ]*(?P<title>.*)$
resize_with:
photos_per_json_file: 500

# Can be daily. daily or monthly.
timeline_granularity: monthly

"""



#!/usr/bin/python
# coding: utf-8

"""Happy Tree Gallery: Static Gallery Generator
   If correctly installed type:
                    htg --help for help.
"""

# import os
import click

from .commands import album
from .commands import config
from .commands import groom
from .commands import index
from .commands import jsongen
from .commands import resize
from .commands import util

from . import VERSION
from . import APP_VERBOSE_NAME

from .utils.aliased_group import AliasedGroup


@click.group(
    cls=AliasedGroup,
    help="%s is a static photo gallery generator." % APP_VERBOSE_NAME,
    short_help="%s is a static photo gallery generator." % APP_VERBOSE_NAME,
    epilog="Use \"htg COMMAND --help\" for help with subcommands.",
)
@click.version_option(
    version=VERSION,
    message='%(version)s',
)
def htg(*args, **kwargs):
    pass


# Add the subcommands
htg.add_command(album)
htg.add_command(config)
htg.add_command(groom)
htg.add_command(index)
htg.add_command(jsongen)
htg.add_command(resize)
htg.add_command(util)


def main():
    htg()



import click
import datetime
import os
import yaml

from .. import APP_VERBOSE_NAME
from ..defaults import DEFAULT_CONFIG, DEFAULT_ADVANCED
from ..models import init_database
from ..models import Size
from ..utils import fatal_error
from ..utils import slugify

COMPULSORY_NOT_NONE = [
    'resized_photos_dir',
    'resized_photos_url',
    'json_data_dir',
    'json_data_url',
]


class HTGConfigError(Exception):
    def __init__(self, msg):
        self.msg = msg


class Config(object):

    errors = []

    def __init__(self, config_file=None, create=False, apply_changes=True):
        self.file = config_file
        if self.file is None:
            self.file = self._default_config_file
        if create:
            return self.create()
        if not os.path.exists(self.file):
            return fatal_error('Config file does not exist.')
        return self.read(apply_changes)

    @property
    def _default_config_file(self):
        return os.path.join(
            click.get_app_dir(APP_VERBOSE_NAME),
            'config.yaml'
        )

    def _check_sizes(self, apply_changes=True):
        sizes = {}
        self.sizes_in_json = []
        for size_name in self.sizes:
            size_dict = {
                'name': slugify(size_name),
                'width': self.sizes[size_name].get('width', None),
                'height': self.sizes[size_name].get('height', None),
                'crop': self.sizes[size_name].get('crop', False),
                'include_in_json': self.sizes[size_name].get(
                    'include_resized_size_in_json',
                    False,
                ),
            }
            if size_dict['height'] is None:
                size_dict['height'] = size_dict['width']
            try:
                size_dict['width'] = int(size_dict['width'])
            except (ValueError, TypeError):
                raise HTGConfigError(
                    'Size with name %s does not have a valid width value.' %
                    self.file
                )
            try:
                size_dict['height'] = int(size_dict['height'])
            except ValueError:
                raise HTGConfigError(
                    'Size with name %s does not have a valid height value.' %
                    self.file
                )

            size = Size.get_size_by_name(size_dict['name'])
            if size is None:
                size = Size(
                    name=size_dict['name'],
                    width=size_dict['width'],
                    height=size_dict['height'],
                    crop=size_dict['crop'],
                    include_in_json=size_dict['include_in_json'],
                )
            else:
                changed = False
                for key in size_dict.keys():
                    if not getattr(size, key) == size_dict[key]:
                        setattr(size, key, size_dict[key])
                        changed = True
                if changed:
                    size.change_date = datetime.datetime.now()
            if apply_changes:
                size.save()
            sizes[size.name] = size
            if apply_changes and size.include_in_json:
                self.sizes_in_json.append(size.id)

        self.sizes = sizes

    def _update_config(self, config):
        for param in config:
            # if isinstance(config[param], dict):
            #     config[param] = Struct(config[param])
            setattr(self, param, config[param])

    def create(self):
        if self.file == self._default_config_file and not os.path.exists(
            os.path.dirname(self.file)
        ):
            try:
                os.makedirs(os.path.dirname(self.file), 0o775)
            except OSError:
                return fatal_error('Could not create the config directory')
        if not os.path.exists(self.file) or click.confirm(
                'The config file already exists. Overwrite?'
        ):
            fp = open(self.file, 'w')
            fp.write(DEFAULT_CONFIG)
            fp.close()
            click.echo("Your config file was saved at: %s" % self.file)
        else:
            click.echo("Your config file could not be saved.")

    def read(self, apply_changes=True):
        base_config = yaml.load(DEFAULT_CONFIG + DEFAULT_ADVANCED)
        self._update_config(base_config)
        stream = open(self.file, 'r')
        try:
            user_config = yaml.load(stream)
        except Exception:
            raise HTGConfigError(
                'Your config file could not be parsed. Perhaps it is not '
                'a valid YAML file. You can check your YAML at: '
                'http://www.yamllint.com/'
            )
        self._update_config(user_config)

        # Set the database file if it isn't set.
        if not self.database_file:
            self.database_file = os.path.join(
                os.path.dirname(self.file),
                'db.sqlite'
            )

        def _check_dir(dir_path, name, writable=True):
            if not isinstance(dir_path, str):
                self.errors.append(('The value for %s is not a valid string.')
                                   % name)
            elif os.path.isdir(dir_path) and not os.access(dir_path, os.W_OK):
                self.errors.append(('The %s cannot be created because the '
                                    'the directory %s is not writable.') %
                                   (name, dir_path))
            elif not os.path.isdir(dir_path):
                self.errors.append(('The %s cannot be created because the '
                                    'the directory %s is not a directory.') %
                                   (name, dir_path))

        # Check if the database is writable or can be created.
        if os.path.exists(self.database_file) and \
                not os.access(self.database_file, os.W_OK):
            self.errors.append('The database_file already exists but it is '
                               'not writable.')
        elif not os.path.exists(self.database_file):
            _check_dir(os.path.dirname(self.database_file), 'database_file')

        _check_dir(self.resized_photos_dir, 'resized_photos_dir')
        _check_dir(self.json_data_dir, 'json_data_dir')

        if self.errors:
            raise HTGConfigError(
                'There were errors with one or more of your paths: \n- %s' %
                '\n- '.join(self.errors)
            )

        # Filter out paths that do not exist.
        self.original_photos_dirs = [
            p for p in self.original_photos_dirs if
            isinstance(p, str) and os.path.isdir(p)
        ]

        # Initialize the database
        init_database(self.database_file)

        # check_file
        for attr in COMPULSORY_NOT_NONE:
            if getattr(self, attr) is None:
                raise HTGConfigError("The value for %s is not set." % attr)

        self._check_sizes(apply_changes)



# coding: utf-8

import click

from ..classes import Config
from ..classes import Resize
from ..classes import ResizeSimple
from ..utils import abort


@click.command('resize', help="Resize pictures to the specified sizes.")
@click.option(
    '--do-not-track',
    help='Normally resized pictures are stored in the database and can be '
         'regenerated when the source file changes. Set this flag '
         'to not track the resizes in the database.',
    is_flag=True,
)
@click.option(
    '--force',
    help='Force resizing of all photos irregardless of their state.',
    is_flag=True,
)
@click.option(
    '--output-dir', '-o',
    help='Set a different output dir. This will overrule the '
         'resized_photos_dir setting of your config file.',
    type=click.Path(exists=True, file_okay=False, writable=True,
                    resolve_path=True)
)
@click.option(
    '--size', '-s',
    help='Specify the size(s) to process when resizing. If no size is '
         'specified all sizes will be processed. Can be set multiple times. '
         'Please note that the size names must match the names in your '
         'config file.',
    multiple=True,
    # callback=validate_resize_size,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def resize(config_file, do_not_track=False, force=False, output_dir=None,
           size=[], *args, **kwargs):
    resize_class = ResizeSimple if do_not_track else Resize
    try:
        resize = resize_class(
            Config(config_file),
            force=force,
            sizes=size,
            output_dir=output_dir,
        )
        resize.start()
    except KeyboardInterrupt:
        return abort()



# coding: utf-8

import click
import os
import yaml

from ..classes import Clean
from ..classes import Config
from ..classes import Index
from ..classes import Serve

from ..meta_example import META_EXAMPLE
# from ..migrate import migrate
from ..models import Album

from ..utils import abort
from ..utils import fatal_error
from ..utils.aliased_group import AliasedGroup


@click.group(
    'utils',
    cls=AliasedGroup,
    help="Various utilities",
    epilog="Use \"htg utils COMMAND --help\" for help with subcommands.",
)
def util(*args, **kwargs):
    pass


@click.command(
    'clean',
    help='Clean database and remove all references to pictures that no '
         'longer exist on your file system. Also deletes all resized pictures '
         'based on the pictures that are removed.',
)
@click.option(
    '--keep-resized',
    help='By default all resized versions of originals that can no longer be '
         'found on your file system are removed from the database and deleted '
         'from your file system. Set this flag to keep these resized pictures '
         'anyway',
    is_flag=True,
)
@click.option(
    '--resize-dir',
    help='Directory in which to look for resized images that are no longer in '
         'the database. Found resized images will be deleted. Defaults to '
         'resized_photos_dir setting from your config file.',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def clean(config_file, keep_resized=False, resize_dir=None, *args, **kwargs):
    try:
        clean = Clean(Config(config_file))
        clean.clean(keep_resized=keep_resized, resize_dir=resize_dir)
    except KeyboardInterrupt:
        return abort()

util.add_command(clean)


@click.command(
    'meta',
    help='Tests the METAFILE if the file exists or attempts to create a meta '
         'file with example data if the file does not exist.',
)
@click.argument(
    'meta-file',
    required=True,
    metavar="METAFILE",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True)
)
def meta(meta_file):
    if os.path.isfile(meta_file):
        index = Index()
        meta_data = index.test_meta(meta_file)
        if meta_data is None:
            click.echo("File could not be parsed.")
        else:
            click.echo("%s" % meta_data)
            if meta_data.get('description'):
                click.echo("Description:\n".upper())
                click.echo("%s" % meta_data['description'])
                del(meta_data['description'])
            click.echo("\nParsed and reformatted data:\n".upper())
            click.echo("%s" % yaml.dump(meta_data, default_flow_style=False))
            click.echo("\nFile is ok.".upper())
    else:
        if not os.path.isdir(os.path.dirname(meta_file)):
            return fatal_error(
                "The directory %s does not exist." % os.path.dirname(meta_file)
            )
        try:
            fp = open(meta_file, 'w')
        except IOError:
            return fatal_error("The file %s is not writable." % meta_file)
        fp.write(META_EXAMPLE)
        fp.close()
        click.echo("Created example meta file at: %s" % meta_file)

util.add_command(meta)


@click.command(
    'move',
    help='Moved changes file paths in the db after you moved files on your '
         'file system.'
)
@click.option(
    '--from-dir', '-f',
    required=True,
    type=click.Path(exists=False, file_okay=False, resolve_path=True),
    help="The old directory"
)
@click.option(
    '--to-dir', '-t',
    required=True,
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="The new directory (needs to exist)."
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def move(config_file=None, from_dir=None, to_dir=None, *args, **kwargs):
    try:
        clean = Clean(Config(config_file))
        clean.move(from_dir=from_dir, to_dir=to_dir)
    except KeyboardInterrupt:
        return abort()

util.add_command(move)


@click.command('reset', help='Delete albums or the entire database. Beware!')
@click.option(
    '--db',
    help='Resets everything by deleting the database file. Otherwise only the '
         'albums will be deleted.',
    is_flag=True,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def reset(config_file, db=False, *args, **kwargs):
    config = Config(config_file)
    if db and click.confirm('Are you sure you want to delete your '
                            'database file?'):
        os.unlink(config.database_file)
    else:
        Album.delete().execute()

util.add_command(reset)


@click.command('serve', help="Simple webserver to test your gallery.")
@click.option(
    '--host',
    help='Select the hostname for your local server.',
    default='localhost',
)
@click.option(
    '--port',
    help='Select the port.',
    default=8000,
    type=int,
)
@click.argument(
    'document-root',
    required=False,
    type=click.Path(exists=True, file_okay=False, resolve_path=True)
)
def serve(document_root=None, host='localhost', port=8000, *args, **kwargs):
    Serve.serve(document_root, host=host, port=port, *args, **kwargs)

util.add_command(serve)


# util.add_command(migrate)



# -*- coding: utf8 -*-

import click
import datetime
import os

from ..models import Picture
from ..models import ResizedPicture
from ..utils import abort
from ..utils import check_and_create_dir
from ..imaging import resize_and_cache


class Resize(object):

    queue = []
    resizes_processed = 0
    page = 0

    def __init__(self, config, force=False, sizes=[], output_dir=None):
        self.config = config
        self.force = force
        self.output_dir = output_dir if output_dir else \
            self.config.resized_photos_dir
        self.sizes = [self.config.sizes[s] for s in self.config.sizes if
                      s in sizes or not sizes]

    def start(self):
        aborted = False
        click.echo("Filling up resize queue... ", nl=False)
        try:
            self._scan_pictures()
        except KeyboardInterrupt:
            return abort("\nCancelled queueing. Nothing changed.")
        click.echo("Done.")

        click.echo("Resizing %s pictures." % len(self.queue))
        try:
            with click.progressbar(self.queue, label="Resized:",
                                   length=len(self.queue),
                                   show_pos=True) as bar:
                for job in bar:
                    self._process_resize_job(job)
                    self.resizes_processed += 1
        except KeyboardInterrupt:
            aborted = True
        if aborted:
            msg = "PLEASE NOTE: Aborted after processing %s of %s resizes.\n" \
                  % (self.resizes_processed, len(self.queue))
        else:
            msg = "Done resizing pictures: "
            msg += "Processed %s resizes." % self.resizes_processed
        click.echo(msg)

    def _get_pictures(self):
        self.page += 1
        return Picture.select().order_by(Picture.id).paginate(self.page, 1000)

    def _scan_pictures(self):
        pics = self._get_pictures()
        while pics.count():
            for pic in pics:
                resizes = {}
                for resize in pic.resized_pictures.where(
                    ResizedPicture.picture_path == self.output_dir
                ):
                    resizes[resize.size.name] = resize
                for size in self.sizes:
                    if size.name not in resizes:
                        self.queue.append((pic, size, None))
                    elif size.name in resizes and (
                        resizes[size.name].resize_date is None or
                        resizes[size.name].resize_date < pic.last_change_indexed
                    ):
                        self.queue.append((pic, size, resizes[size.name]))
                    elif self.force:
                        self.queue.append((pic, size, resizes[size.name]))
            pics = self._get_pictures()

    def _process_resize_job(self, job):
        picture, size, resize = job
        if resize is None:
            resize = ResizedPicture.create(
                picture=picture,
                size=size,
                picture_path=self.output_dir,
            )
        resize.resize_date = datetime.datetime.now()
        resize_dir, name = os.path.split(resize.path)
        image_size = False
        if check_and_create_dir(resize_dir, create=True):
            image_size, msg = resize_and_cache(
                self.config.resize_with,
                picture.file_path,
                resize.path,
                size.width,
                size.height,
                crop=size.crop,
                rotation=picture.rotation,
            )
        if image_size:
            resize.width = image_size[0]
            resize.height = image_size[1]
            resize.successfully_created = True
        resize.save()



# coding: utf-8

import os
import re
import shutil
from decimal import Decimal, ROUND_UP

from ..utils import execute

SIPS = '/usr/bin/sips'
HAS_JPEGTRAN, msg = execute(['which', 'jpegtran'])


def _get_dimensions(path):
    def _get_wh(keyword):
        m = re.search(r'%s: (\d+)' % keyword, output)
        if m:
            return int(m.group(1))
        return 0
    success, output = execute(
        [SIPS, '-g', 'pixelWidth', '-g', 'pixelHeight', path]
    )
    if success:
        return _get_wh('pixelWidth'), _get_wh('pixelHeight')
    return 0, 0


def resize_and_cache(org_file, new_file, width, height,
                     crop=False, rotation=0):
    def _failed(msg):
        os.unlink(new_file)
        return False, msg

    try:
        shutil.copy(org_file, new_file)
    except IOError:
        return _failed('Could not copy the original file to the new location.')

    # if the image has to fix within a square bounding box, it's easy with sips
    if width == height and not crop:
        if rotation:
            success, output = execute([SIPS, '-Z', width, '-r', rotation, new_file])
        success, output = execute([SIPS, '-Z', width, new_file])
        if not success:
            return _failed('Something went wrong while executing SIPS.')
        return _get_dimensions(new_file), False

    # Rotate first
    if rotation:
        # We prefer jpegtran, because it's rotations are lossless.
        if rotation in [90, 180, 270] and HAS_JPEGTRAN:
            success, output = execute([
                'jpegtran',
                '-rotate',
                rotation,
                '-outfile',
                new_file,
                new_file
            ])
        elif rotation:
            success, output = execute([SIPS, '-r', rotation, new_file])
        if not success:
            return _failed('Could not rotate the image.')

    # Retreive the orignal dimensions
    orgW, orgH = _get_dimensions(new_file)
    if not orgW or not orgH:
        return _failed('Could not get original dimensions of the image file.')

    # Calculate the transformations
    arorg = Decimal(orgW) / Decimal(orgH)
    arnew = (Decimal(width) / Decimal(height))

    new_w, new_h = False, False

    if crop:
        # the original is taller
        if arorg < arnew:
            # first resize to width
            new_w = width
            new_h = int((Decimal(width) / arorg).quantize(
                Decimal('1'),
                rounding=ROUND_UP
            ))
        # The orginal image is wider
        elif arorg > arnew:
            new_h = height
            new_w = int((Decimal(height) * arorg).quantize(
                Decimal('1'),
                rounding=ROUND_UP
            ))
        # Do nothing if the original picture has the same aspect ratio

    # check if the image is larger or not then the image we want to resize to.
    elif orgW > width or orgH > height:
        # The original image is taller
        if arorg < arnew:
            new_w = int((Decimal(height) * arorg).quantize(
                Decimal('1'),
                rounding=ROUND_UP
            ))
            new_h = height
        # The orginal image is wider
        elif arorg > arnew:
            new_w = width
            new_h = int((Decimal(width) / arorg).quantize(
                Decimal('1'),
                rounding=ROUND_UP
            ))

    if new_w and new_h:
        success, output = execute([SIPS, '-z', new_h, new_w, new_file])
        if not success:
            return _failed('Could not resize image.')

    if crop:
        success, output = execute([SIPS, '-c', height, width, new_file])
        if not success:
            return _failed('Could not crop the image.')

    return _get_dimensions(new_file), False



# -*- coding: utf8 -*-

import click
import datetime
import os
import peewee
import pickle
import warnings

from ..imaging import image_metadata
from ..models import Album
from ..models import Picture
from ..utils import abort
from ..utils import checksum
from ..utils import fatal_error
from ..utils import file_name_date
from ..utils import filter_image_files
from ..utils import parse_meta_file
from ..utils import title_and_date_from_dir_name


META_FIELDS = ['title', 'start_date', 'end_date', 'description', 'meta']


class Index(object):

    new, processed, changed, unchanged, duplicates = 0, 0, 0, 0, 0
    checksums = []
    photos = []
    failed_meta_files = []

    def __init__(self, config=None):
        self.config = config

    def _meta_from_path(self, path):
        meta = {}
        meta_file = os.path.join(path, self.config.meta_file_name)
        if os.path.exists(meta_file):
            meta = parse_meta_file(meta_file)
            if meta is None:
                self.failed_meta_files.append(meta_file)
        return {} if meta is None else meta

    def _meta_from_dir_name(self, path, meta={}):
        if meta.get('title') and meta.get('start_date'):
            return meta
        title, start_date = title_and_date_from_dir_name(
            path, self.config.album_dir_pattern
        )
        if title and not meta.get('title', False):
            meta['title'] = title
        if start_date and not meta.get('start_date', False):
            meta['start_date'] = start_date
        return meta

    def _album_from_meta(self, path, meta):
        try:
            album = Album.get(album_path=path)
        except peewee.DoesNotExist:
            album = None
        if not meta or meta.get('album') is False:
            if album:
                album.delete_instance(recursive=True)
            return None
        if meta.get('album'):
            key = meta['album'].split('-')[-1].lower()
            try:
                album = Album.get(key=key)
            except peewee.DoesNotExist:
                pass
        # Pickling in protocol 0 and converting to string for Python 2 compatibility
        meta['meta'] = pickle.dumps(meta.get('meta', {}), 0).decode()
        if album is None and meta.get('title'):
            album = Album.create(
                album_path=path,
                start_date=meta.get('start_date'),
                end_date=meta.get('end_date'),
                title=meta['title'],
                description=meta.get('description'),
                meta=meta['meta'],
            )
        else:
            updated = False
            for f in META_FIELDS:
                if meta.get(f) is not None and not getattr(album, f) == meta[f]:
                    setattr(album, f, meta[f])
                    updated = True
            if updated:
                album.save()
        return album

    def _echo_failed_meta_files(self):
        if self.failed_meta_files:
            click.echo(79 * "-")
            click.echo(
                "Please Note: "
                "The YAML parser could not parse the following meta files:"
                "\n- %s" % "\n- ".join(self.failed_meta_files)
            )
            click.echo(79 * "-")

    def _process_single_photo(self, photo, reprocess_image_meta=False):
        path, file_name, album = photo
        file_path = os.path.join(path, file_name)
        md5sum = checksum(file_path)
        changed, duplicate = False, False

        # Detect a duplicate picture in a different place.
        if md5sum in self.checksums:
            self.duplicates += 1
            duplicate = True
        else:
            self.checksums.append(md5sum)

        try:
            pic = Picture.get(checksum=md5sum)
        except peewee.DoesNotExist:
            try:
                pic = Picture.get(file_path=file_path)
                pic.checksum = md5sum
                changed = True
            except peewee.DoesNotExist:
                pic = None

        # Check for changes in existing pic.
        if pic is not None:
            with warnings.catch_warnings():  # For comparing apples and pears
                warnings.simplefilter("ignore")
                if not pic.file_path == file_path and not duplicate:
                    pic.file_path = file_path
                if not pic.file_name == file_name and not duplicate:
                    pic.file_name = file_name
                    changed = True  # Because date in filename could've changed

        if pic is None or changed or reprocess_image_meta:
            meta = image_metadata(file_path)
            if meta['date'] is None:
                meta['date'] = file_name_date(
                    file_path,
                    self.config.file_date_formats,
                    self.config.use_file_creation_date_as_image_date,
                )

        if pic is None and not duplicate:
            self.new += 1
            pic = Picture.create(
                checksum=md5sum,
                file_path=file_path,
                file_name=file_name,
                sort_date=meta['date'],
                caption=meta['caption'],
                latitude=meta['gps'][0],
                longitude=meta['gps'][1],
                rotation=meta['rotation'],
            )
        elif changed or reprocess_image_meta:
            self.changed += 1
            pic.sort_date = meta['date']
            pic.caption = meta['caption']
            pic.latitude = meta['gps'][0]
            pic.longitude = meta['gps'][1]
            pic.rotation = meta['rotation']
            if changed:
                pic.last_change_indexed = datetime.datetime.now()
            pic.save()
        else:
            self.unchanged += 1

        if album is not None and album not in pic.albums:
            pic.albums.add(album)

        self.processed += 1
        return pic

    def index(self, dirs=[], reprocess_image_meta=False):
        photo_dirs = dirs if dirs else self.config.original_photos_dirs
        if not photo_dirs:
            return fatal_error("There were no directories to index.")
        click.echo("Finding all photos and directories in: ")
        excluded_paths = []
        album_paths = {}
        try:
            for nr, pd in enumerate(photo_dirs, start=1):
                click.echo("%s. %s" % (nr, pd))
                for path, dirs, files in os.walk(pd):
                    # See if this path should be included.
                    if path == pd:
                        continue
                    if list(filter(lambda p: path.startswith(p), excluded_paths)):
                        continue
                    if os.path.exists(os.path.join(
                        path, self.config.exclude_file_name
                    )):
                        excluded_paths.append(path)
                        continue

                    click.echo("  - %s" % path.replace(pd + '/', ''))
                    # Check if this path is inside an existing album
                    album = None
                    for p in album_paths:
                        if path.startswith(p):
                            album = album_paths[p]
                            short_path = path.replace(p + '/', '')
                            break
                    if album is None:
                        # Check for available meta data
                        meta = self._meta_from_path(path)
                        if self.config.create_albums_from_dir_names:
                            meta = self._meta_from_dir_name(path, meta)

                        # If there is meta data, see which album it goes into.
                        album = self._album_from_meta(path, meta.copy())

                        if album:
                            if isinstance(
                                meta.get('meta', {}).get('exclude'), list
                            ):
                                album.exclude = meta['meta']['exclude']
                            album_paths[path] = album
                            short_path = None
                    # Index files if they should be included
                    if album and short_path in album.exclude:
                        continue
                    for f in filter_image_files(self.config, files):
                        self.photos.append((path, f, album))
        except KeyboardInterrupt:
            return abort(
                "Cancelled during indexing. Meta data may have been updated."
            )
        click.echo("All directories have been indexed.\n")

        self._echo_failed_meta_files()

        aborted = False
        click.echo("Processing %s photos." % len(self.photos))
        try:
            with click.progressbar(self.photos, label="Processed:",
                                   length=len(self.photos),
                                   show_pos=True) as bar:
                for photo in bar:
                    self._process_single_photo(
                        photo,
                        reprocess_image_meta=reprocess_image_meta,
                    )
        except KeyboardInterrupt:
            aborted = True
        if aborted:
            msg = "PLEASE NOTE: Aborted after processing %s of %s photos.\n" \
                  % (self.processed, len(self.photos))
        else:
            msg = "Done processing photos:\n"
        for name in ['new', 'changed', 'duplicates', 'unchanged']:
            count = getattr(self, name)
            msg += "{name:>12}: {nr:>5}\n".format(name=name, nr=count)
        msg += "{name:>12}  {nr:>5}\n".format(name="", nr="————— +")
        msg += "{name:>12}: {nr:>5}\n".format(name="total",
                                              nr=self.processed)
        click.echo(msg)

    def update_album_meta(self):

        def _get_albums(page=0):
            page += 1
            return Album.select().order_by(Album.id).paginate(page, 1000), page
        click.echo("Checking for updated meta file in existing albums.")
        total = Album.select().order_by(Album.id).count()
        albums, page = _get_albums()
        with click.progressbar(length=total, label='Albums:',
                               show_pos=True) as bar:
            click.echo
            while albums.count():
                for album in albums:
                    bar.update(1)
                    meta = self._meta_from_path(album.album_path)
                    if not meta:
                        continue
                    # Pickling in protocol 0 and converting to string for Python 2 compatibility
                    meta['meta'] = pickle.dumps(meta.get('meta', {})).decode()
                    updated = False
                    for p in META_FIELDS:
                        if meta.get(p) is not None and not \
                                getattr(album, p) == meta.get(p):
                            setattr(album, p, meta[p])
                            updated = True
                    if updated:
                        album.save()
                albums, page = _get_albums(page)

        self._echo_failed_meta_files()
        click.echo("Done.")

    def test_meta(self, meta_file):
        return parse_meta_file(meta_file)



import click
import os
import yaml

from ..utils import parse_meta_file


def in2ex(config, dirs):
    skip_dirs = []
    click.echo("Updating meta files")
    for root_dir in dirs:
        for path, dirs, files in os.walk(root_dir):
            if path == root_dir:
                continue
            if list(filter(lambda d: path.startswith(d), skip_dirs)):
                continue
            exclude_file = os.path.join(path, config.exclude_file_name)
            meta_file = os.path.join(path, config.meta_file_name)
            if os.path.exists(exclude_file):
                skip_dirs.append(path)
                continue
            if not os.path.exists(meta_file):
                continue
            skip_dirs.append(path)
            meta, description = parse_meta_file(meta_file, raw=True)
            include = meta.get('include', None)
            if include:
                del(meta['include'])
            if not isinstance(include, list):
                include = []
            exclude = exclude_subdirs(path, include)
            if not exclude:
                continue
            meta['exclude'] = exclude
            if meta.get('meta', None) is not None:
                for m in meta['meta']:
                    meta[m] = meta['meta'][m]
                del(meta['meta'])
            meta_yaml = yaml.dump(meta, default_flow_style=False)
            meta_yaml = meta_yaml.replace('\n-', '\n  -')
            if description is not None:
                meta_yaml = "%s\n---\n%s" % (meta_yaml, description)
            try:
                fp = open(meta_file, 'w')
                fp.write(meta_yaml)
                fp.close()
                click.echo("Updated: %s" % meta_file)
            except OSError:
                click.echo("ERROR: %s" % meta_file)
                click.echo("       Meta file not updated, because it could "
                           "not be overwritten.")
    click.echo("Done.")


def exclude_subdirs(album_path, include=[]):
    exclude = []
    for path, dirs, files in os.walk(album_path):
        if path == album_path:
            continue
        dir_name = path.replace(album_path + '/', '')
        if dir_name not in include:
            exclude.append(dir_name)
    exclude.sort()
    return exclude

    click.echo("Done")



# coding: utf-8

import mimetypes
import os
import socket
import sys
import webbrowser

from socketserver import ThreadingMixIn
from socketserver import TCPServer
from http.server import SimpleHTTPRequestHandler

from ..utils import fatal_error


class Serve(object):

    @staticmethod
    def serve(document_root=None, host='localhost', port='8000', *args,
              **kwargs):
        if document_root is not None:
            os.chdir(document_root)
        try:
            httpd = SimpleServe((host, port), SimpleRequests)
        except socket.error:
            return fatal_error(
                'Could not open the webserver at the requested (%s) port' % port
            )
        webbrowser.open('http://%s:%s' % (host, port))
        try:
            httpd.serve_forever()
        except (KeyboardInterrupt, SystemExit):
            print('Done serving. Goodbye.')
            httpd.server_close()


class SimpleServe(ThreadingMixIn, TCPServer):
    allow_reuse_address = True


class SimpleRequests(SimpleHTTPRequestHandler):

    def send_head(self):
        path = self.translate_path(self.path)

        f = None
        if os.path.isdir(path):
            if not self.path.endswith('/'):
                # redirect browser - doing basically what apache does
                self.send_response(301)
                self.send_header("Location", self.path + "/")
                self.end_headers()
                return None
            for index in "index.html", "index.htm":
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            # else:
            #   return self.list_directory(path)

        try:
            # Always read in binary mode. Opening files in text mode may cause
            # newline translations, making the actual size of the content
            # transmitted *less* than the content-length!
            print("%s" % path)
            f = open(path, 'rb')

        except IOError:
            self.send_error(404, "File not found")
            return None

        fs = os.fstat(f.fileno())
        tp = self.guess_type(path)

        headers = {
            "Content-type": tp,
            "Content-Length": str(fs[6]),
            "Cache-Control": "no-cache, must-revalidate",
        }

        # Last-Modified", self.date_time_string(fs.st_mtime)

        return self.send_content(200, headers, f)

    def send_content(self, code, headers, file_handler):

        self.send_response(code)

        for key, value in headers.iteritems():
            self.send_header(key, value)

        self.end_headers()

        return file_handler

    def log_message(self, format, *args):
        sys.stdout.write("%s\n" % format % args)

    def log_request(self, code='', size=''):
        try:
            self.log_message(
                '%s %s %s',
                str(code),
                self.requestline.split(' ')[0],
                self.requestline.split(' ')[1]
            )
        except:
            pass

    def guess_type(self, path):
        mimetype, encoding = mimetypes.guess_type(path)
        return mimetype if mimetype else 'text/html'



# coding: utf-8

from ..utils import execute


def _get_dimensions(path):
    success, output = execute(
        ['convert', path, '-ping', '-format', '%wx%h', 'info:']
    )
    if not success:
        return 0, 0
    return [int(i) for i in output.split('x')]


def resize_and_cache(org_file, new_file, width, height,
                     crop=False, rotation=0):
    command = ['convert', org_file]
    geometry = '%sx%s' % (width, height)
    if rotation:
        command += ['-rotate', rotation]
    if crop:
        command += [
            '-resize',
            geometry + '^',
            '-gravity',
            'center',
            '-extent',
            geometry,
        ]
    else:
        command += [
            '-resize',
            geometry + '>',
        ]
    command.append(new_file)
    success, msg = execute(command)
    if success:
        return _get_dimensions(new_file), False
    else:
        return False, 'Could not convert image.'



import click

from ..classes import Config
from ..utils.groom import groom as groomfunc


@click.command(
    'groom',
    help='Groom your photo directories by walking through them and adding '
         'a meta or exclusion file for better indexing',
)
@click.option(
    '--include-all', '-a',
    help='By default directories that already have an exclusion or meta_file '
         'will be skipped. Setting this option will also force you to '
         'decide again.',
    is_flag=True,
)
@click.option(
    '--directory', '-d',
    required=False,
    metavar="DIR",
    multiple=True,
    help='The grooming will default to the original_photos_dirs setting '
         'in the config file if you do not provide root directories.',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def groom(config_file=None, directory=[], include_all=False):
    config = Config(config_file)
    dirs = directory if directory else config.original_photos_dirs
    groomfunc(config, dirs, include_all=include_all)



import click

from ..classes import Config
from ..classes import Index


@click.command('index', help="Index all photos in your source directory")
@click.option(
    '--index-dir', '-i',
    help='Specify directory where your original photos are. Can be set '
         'multiple times.',
    multiple=True,
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)
@click.option(
    '--update-album-meta', '-u',
    help='Only check existing meta files and save changes. No photo or '
         'directory indexing.',
    is_flag=True,
)
@click.option(
    '--image-meta', '-m',
    help='Reprocess all image meta data.',
    is_flag=True,
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def index(config_file, index_dir=[], image_meta=False, update_album_meta=False,
          *args, **kwargs):
    index = Index(Config(config_file))
    if update_album_meta:
        index.update_album_meta()
    else:
        index.index(dirs=index_dir, reprocess_image_meta=image_meta)



from .clean import Clean
from .config import Config
from .config import HTGConfigError
from .index import Index
from .json import Json
from .resize import Resize
from .resize_simple import ResizeSimple
from .serve import Serve



# coding: utf-8

import click

from ..models import Size


def validate_resize_size(ctx, param, name):
    def _validate(name):
        if Size.get_size_by_name(name):
            return name
        raise click.BadParameter(
            '%s is not a valid size name. Please choose from: \n  - %s' %
            (name, "\n  - ".join(Size.all_names()))
        )
    if hasattr(name, '__iter__'):
        return [_validate(n) for n in name]
    else:
        return _validate(name)



import click
import datetime
import os
import sys

from ..meta_example import META_EXAMPLE_SHORT
from ..meta_example import META_EXAMPLE_NO_ALBUM
from ..utils import filter_image_files
from ..utils import title_and_date_from_dir_name
from ..utils import user_input


def groom_subdirs(config, album_path):
    exclude, sub_dirs = [], []
    click.echo('  Grooming subdirectories of the album:')
    for path, dirs, files in os.walk(album_path):
        path = '' if path == album_path else path.replace(album_path + '/', '')
        if filter_image_files(config, files):
            sub_dirs.append(path)
        else:
            exclude.append(path)
    # Reversing so we start with the deepest directories first. This makes
    # more sense when grooming.
    sub_dirs.reverse()
    for d in sub_dirs:
        text = 'Album root' if '' == d else d
        click.echo('  - %s:' % text, nl=False)
        if not click.confirm(' Include contents in album?'):
            exclude.append(d)
    if exclude:
        exclude.sort()
        return "exclude:\n  - %s" % "\n  - ".join(exclude)
    return ''


def groom(config, dirs, include_all=False):
    click.echo(
        'Groom is going to loop through all your directories and for each '
        'directory you will get 4 choices: \n'
        '\n'
        '- i: Ignore the directory\n'
        '- x: Exclude the directory (1)\n'
        '- a: Include the directory and consider its contents as an album (2)\n'
        '- n: Include the media in the directory, but it\'s not an album (2)\n'
        '\n'
        '(1): An exclusion file will be added to the directory. The name of '
        'this file can be set with the "exclude_file_name" '
        'setting in your config file.\n'
        '(2): A meta file will be added to the directory. The name of this '
        'meta file is set in the "meta_file_name" setting inside the config '
        'file.\n'
    )
    if not click.confirm('All clear?'):
        click.echo('Ok, aborting..')
        sys.exit()
    skip_dirs = []
    for root_dir in dirs:
        click.echo("Grooming directory: %s" % root_dir)
        for path, dirs, files in os.walk(root_dir):
            if path == root_dir:
                continue
            if list(filter(lambda d: path.startswith(d), skip_dirs)):
                continue
            click.echo('- %s' % path.replace(root_dir, ''))
            exclude, meta = False, False
            exclude_file = os.path.join(path, config.exclude_file_name)
            meta_file = os.path.join(path, config.meta_file_name)
            if os.path.exists(exclude_file):
                exclude = exclude_file
            if os.path.exists(meta_file):
                meta = meta_file
            if exclude and not include_all:
                skip_dirs.append(path)
                click.echo('  Skipped: Already excluded.')
                continue
            if meta and not include_all:
                skip_dirs.append(path)
                click.echo('  Skipped: Already has a meta file.')
                continue
            if not os.access(path, os.W_OK):
                click.echo(
                    " Skipped: Directory is not writable."
                )
                continue
            answer = None
            while answer not in ['i', 'x', 'a', 'n']:
                answer = user_input(
                    # 'Ignore, eXclude, Album, or Not an album? '
                    '  [i|x|a|n]: '
                )
            if answer == 'x':
                if exclude:
                    # click.echo('  Already excluded, nothing changed')
                    continue
                if meta and not click.confirm(
                    '  This directory currently has a meta file.\n'
                    '  Are you sure you want to delete this file?'
                ):
                    click.echo('  Skipped')
                    continue
                if meta:
                    try:
                        os.unlink(meta)
                    except OSError:
                        click.echo('  ERROR: could not delete the meta file.')
                        continue
                fp = open(exclude_file, 'w')
                fp.close()
                skip_dirs.append(path)
                continue

            if answer == 'a' or answer == 'n':
                if exclude:
                    try:
                        os.unlink(exclude)
                    except OSError:
                        click.echo(
                            '  ERROR: could not delete the exclusion file.'
                        )
                        continue
                if meta and not click.confirm(
                    '  This directory already has a meta file.\n'
                    '  Are you sure you want to overwrite this file?'
                ):
                    click.echo('  Skipped')
                    continue
                if answer == 'a':
                    title, start_date = title_and_date_from_dir_name(
                        path, config.album_dir_pattern
                    )
                    if title is None:
                        title = "Title of your Album"
                    if start_date is None:
                        start_date = datetime.datetime.now()
                    extra = groom_subdirs(config, path) if dirs else ''
                    content = META_EXAMPLE_SHORT % {
                        'title': title,
                        'start_date': start_date.strftime("%Y-%m-%d"),
                        'extra': extra,
                    }
                else:
                    content = META_EXAMPLE_NO_ALBUM
                fp = open(meta_file, 'w')
                fp.write(content)
                fp.close()
                skip_dirs.append(path)
                continue

            # ignoring
            if exclude:
                try:
                    os.unlink(exclude)
                except OSError:
                    click.echo(
                        '  ERROR: could not delete the exclusion file.'
                    )
                    continue
            if meta and not click.confirm(
                '  This directory currently has a meta file.\n'
                '  Are you sure you want to delete this file?'
            ):
                click.echo('  Skipped')
                continue
            if meta:
                try:
                    os.unlink(meta)
                except OSError:
                    click.echo('  ERROR: could not delete the meta file.')
                    continue
    click.echo("Done")



# coding: utf-8

import datetime
import json
import sys

from ..utils import execute
from ..utils import fatal_error
from .resize_im import resize_and_cache as resize_with_im
from .resize_sips import resize_and_cache as resize_with_sips

try:
    from resize_pil import resize_and_cache as resize_with_pil
except ImportError:
    resize_with_pil = False

EXIFTOOL_AVAILABLE, output = execute(['which', 'exiftool'])

EXIF_TITLE_TAGS = [
    'ImageDescription',
    'UserComment',
    'Caption-Abstract',
    'Headline',
]

EXIF_ROTATION_MAP = {
    1: 0,  # Horizontal (normal)
    2: 0,  # Mirrored horizontal
    3: 180,  # Rotated 180
    4: 0,  # Mirrored vertical
    5: 270,  # Mirrored horizontal then rotated 90 CCW
    6: 90,  # Rotated 90 CCW
    7: 90,  # Mirrored horizontal then rotated 90 CW
    8: 270,  # Rotated 90 C
}


def resize_and_cache(resize_with, *args, **kwargs):
    if not resize_with:
        # Based on performance. IM should only be set specifically
        # success, msg = execute(['convert', '--version'])
        # if success and 'ImageMagick' in msg:
        #     preference = 'im'
        if 'darwin' in sys.platform:
            resize_with = 'sips'
        else:
            resize_with = 'pil'

    if resize_with == 'im':
        return resize_with_im(*args, **kwargs)
    elif resize_with == 'sips':
        return resize_with_sips(*args, **kwargs)
    else:
        if resize_with_pil:
            return resize_with_pil(*args, **kwargs)
        else:
            fatal_error('The Python Imaging Library cannot be found.')


def image_caption(raw_data, title_tag=False):
    if title_tag and raw_data.get(title_tag, False):
        return raw_data[title_tag].strip()
    for tag in EXIF_TITLE_TAGS:
        if raw_data.get(tag, False):
            return raw_data[tag].strip()
    return None


def image_date(raw_data):
    tstr = raw_data.get('DateTimeOriginal', None)
    if tstr is None:
        return None
    try:
        return datetime.datetime.strptime(tstr, "%Y:%m:%d %H:%M:%S")
    except ValueError:  # Somehow the format is wrong or all zero's
        return None


def image_geolocation(raw_data):
    if raw_data.get('GPSLongitude', False) and \
            raw_data.get('GPSLatitude', False):
        return [
            raw_data['GPSLatitude'],
            raw_data['GPSLongitude'],
        ]
    return (None, None)


def image_rotation(raw_data):
    orientation = raw_data.get('Orientation', 1)
    return EXIF_ROTATION_MAP.get(orientation, 0)


def image_exif_data(img_path):
    exif_data = {}
    if EXIFTOOL_AVAILABLE:
        success, raw_json = execute(['exiftool', '-j', '-n', img_path])
        if success:
            try:
                return json.loads(raw_json)[0]
            except ValueError:
                # Invalid json input. Let's pass for now
                pass
    return exif_data


def image_metadata(img_path, title_tag=False):
    raw_data = image_exif_data(img_path)
    return {
        'caption': image_caption(raw_data, title_tag),
        'rotation': image_rotation(raw_data),
        'raw_data': raw_data,
        'gps': image_geolocation(raw_data),
        'date': image_date(raw_data)
    }



# -*- coding: utf8 -*-

import click

from ..classes import Config
from ..classes import HTGConfigError
from ..utils import fatal_error
from ..utils.aliased_group import AliasedGroup


@click.group(
    cls=AliasedGroup,
    help="Create and check your config file",
    epilog="Use \"htg config COMMAND --help\" for help with subcommands.",
)
def config(*args, **kwargs):
    pass


@click.command('create', help="Create a new config file.")
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=False, dir_okay=False, resolve_path=True,
                    writable=True),
)
def _create(config_file, *args, **kwargs):
    Config(config_file, create=True)

config.add_command(_create)


@click.command('test', help="Test your config file.")
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True,
                    writable=True),
)
def _test(config_file, *args, **kwargs):
    try:
        Config(config_file, apply_changes=False)
    except HTGConfigError as e:
        return fatal_error(e.msg)
    click.echo("All OK!")

config.add_command(_test)



# coding: utf-8

""" Collection of helper functions that can be used
    in various modules of this program.
"""

import calendar
import click
import datetime
import errno
import hashlib
import json
import markdown2
import os
import random
import re
import string
import subprocess
import sys
import unicodedata
import yaml

from tkinter.filedialog import askdirectory
from tkinter import Tk


def abort(msg="Process was aborted"):
    click.echo("\nPLEASE NOTE: %s" % msg)


def alphanum_hash(length=10):
    """ Generate a random alphanumeric string.
    """
    length = max(0, length)
    population = (string.ascii_letters + string.digits) * length
    return (''.join(random.sample(population, length))).lower()


def ask_for_dir(msg='Pick your directory.'):
    """ Asks the user to input the path.
        The user can choose to open a file dialog.
    """
    directory = user_input("%s\n# " % msg, case_sensitive=True)
    if directory:
        return directory
    return pick_directory()


def check_and_create_dir(path, create=False):
    """ Check if a directory exists and if not
        it can create the directory.
    """
    if os.path.isdir(path):
        return True
    if create:
        try:
            os.makedirs(path, 0o775)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise
        return True
    return False


def execute(command_list):
    """ Executes commands on the command line.
        Returns tuple with the result True|False and
        the output that the command generated.
    """
    try:
        command = [
            c if isinstance(c, str) else str(c) for c in command_list
        ]
        proc = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        out, err = proc.communicate()
        if proc.returncode:
            return False, err.decode('utf-8') if err else None
        return True, out.decode('utf-8') if out else None
    except OSError as e:
        if isinstance(e, FileNotFoundError):  # No such file or directory
            return False, '"%s" is not available on your system.' % command[0]
        else:
            raise e


def fatal_error(message):
    """ Return a fatal error message with description and exit
        the program.
    """
    click.echo("\nFATAL ERROR:")
    sys.exit("%s\n" % message)
    return False


def file_name_date(file_path, date_formats, use_creation_date):
    bn = os.path.basename(file_path)
    date_formats = [
        (df, len(datetime.datetime.now().strftime(df)))
        for df in date_formats
    ]
    for df in date_formats:
        try:
            dt = datetime.datetime.strptime(bn[:df[1]], df[0])
            return dt
        except ValueError:
            pass
    if use_creation_date:
        print(file_path)
        return os.path.getctime(file_path)
    return None


def filter_image_files(config, files):
    images = []
    for f in files:
        name, ext = os.path.splitext(f)
        if ext.lower() in config.image_extensions:
            images.append(f)
    return images


def checksum(file_path, digest="md5", block_size=512):
    digest = getattr(hashlib, digest)()
    f = open(file_path)
    with open(file_path, 'rb') as f:
        for chunk in iter(lambda: f.read(128 * block_size), b''):
            digest.update(chunk)
    return digest.hexdigest()


def parse_json(json_string):
    try:
        return json.loads(json_string)
    except ValueError:
        return False


def parse_meta_file(meta_file, raw=False):
    data, description, meta = None, None, {}

    fp = open(meta_file, 'r')
    raw_meta = fp.read()
    fp.close()

    raw_meta = re.split(r'^--+$', raw_meta, maxsplit=1, flags=re.M)

    # Parse yaml
    data = yaml.load(raw_meta[0])

    if len(raw_meta) > 1:
        description = raw_meta[1]

    if raw:
        return data, description

    if description is not None:
        extras = []
        if data.get('markdown_extras', []):
            extras = data['markdown_extras']
            del(data['markdown_extras'])
        description = markdown2.markdown(description, extras=extras)
        meta['description'] = description
    if data is not None:
        for p in ['title', 'start_date', 'end_date', 'include', 'album']:
            if data.get(p, None) is not None:
                meta[p] = data[p]
                del(data[p])
        if len(data):
            meta['meta'] = data
    return meta


def pick_directory():
    """ Present a UI dialog to easily pick a directory.
    """
    # we don't want a full GUI, so keep the root window from appearing
    Tk().withdraw()
    directory = askdirectory(mustexist=True)
    return directory


def read_json(path):
    """ Read json from a file and return a python dictionary
        Returns None if the file cannot be found/opened
        Returns False if the file does not contain valid json
    """
    try:
        fp = open(path, 'r')
    except IOError:
        return None
    try:
        dictionary = json.load(fp)
    except ValueError:
        return False
    fp.close()
    return dictionary


def slugify(name):
    """
    Converts to lowercase, removes non-word characters (alphanumerics and
    underscores) and converts spaces to hyphens. Also strips leading and
    trailing whitespace.
    """
    # Create unicode sequences instead of characters
    name = unicodedata.normalize('NFKD', name)
    # Replace non letter and space characters
    slug = re.sub('[^\w\s-]', '', name).strip().lower()
    # Replace spaces with -
    slug = re.sub('[-\s]+', '-', slug)
    # Encode and decode to ascii to remove all none ascii characters
    return slug.encode('ascii', 'ignore').decode()


def source_warning(file):
    if os.path.exists(os.path.join(os.path.dirname(file), "..", "README.md")):
        click.echo("\n" + (27 * ">") + " RUNNING FROM SOURCE FILE " + (27 * "<") + "\n")


def title_and_date_from_dir_name(path, pattern):
    title, start_date = None, None
    dir_pattern = re.compile(pattern)
    dir_name = os.path.basename(path)
    match = re.match(dir_pattern, dir_name)
    if match:
        try:
            start_date = datetime.datetime(
                int(match.group('year')),
                int(match.group('month')),
                int(match.group('day')),
            )
        except (IndexError, ValueError):
            # No valid date info could be found or the info was invalid.
            start_date = None
        try:
            title = match.group('title')
        except IndexError:
            title = None
    return title, start_date


def to_utc_timestamp(dt):
    if dt is None:
        return None
    return calendar.timegm(dt.timetuple())


def user_input(msg, case_sensitive=False):
    """ Ask user for input on the command line.
        Returns stripped string of input.
    """
    confirm = (input(u"%s" % msg)).strip()
    if case_sensitive:
        return confirm
    return confirm.lower()


def write_json(path, dictionary):
    """ Transform a python dictionary to JSON.
        Returns True if write is completed successfully
        Returns False if object could not be serialized
        Returns None if file could not be opened

        Writes to a temporary file first and moves the
        file to final location.
    """
    tmp_path = path + '.tmp'
    try:
        fp = open(tmp_path, 'w')
    except IOError:
        return None
    try:
        json.dump(
            dictionary,
            fp,
            indent=2
        )
    except TypeError:
        fp.close()
        return False
    fp.close()
    os.rename(tmp_path, path)
    return True



from setuptools import find_packages, setup

with open('README.md', 'r') as f:
    long_description = f.read()

version = __import__('htg').VERSION

setup(
    name='htg',
    version=version,
    author='R Meurders',
    author_email='pypi+htg@rmnl.net',
    license='MIT',
    description='Happy Tree Gallery is a command line utility generating '
                'static photo galleries.',
    long_description=long_description,
    keywords='development command line tool photos pictures management '
             'galleries json',
    url='https://gitlab.com/rmnl/htg',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click>=7.0,<=7.999',
        'markdown2>=2.3.6,<=2.3.999',
        'numpy>=1.15.2,<=1.15.999',
        'peewee>=3.7.1,<=3.7.999',
        'Pillow>=5.3.0,<=5.3.999',
        'PyYAML>=3.13,<=3.999',
    ],
    entry_points={
        'console_scripts': [
            'htg = htg.__main__:main'
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],
)



import click

from ..classes import Config
from ..classes import Json


@click.command('json', help="Build the json files for your gallery.")
@click.option(
    '--albums',
    help='Only export the json files for the albums view.',
    is_flag=True,
)
@click.option(
    '--timeline',
    help='Only export the json files for the timeline view.',
    is_flag=True,
)
@click.option(
    '--output-dir', '-o',
    help='Set a different output dir. This will overrule the '
         'json_data_dir setting of your config file.',
    type=click.Path(exists=True, file_okay=False, writable=True,
                    resolve_path=True)
)
@click.argument(
    'config-file',
    required=False,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
def jsongen(config_file, output_dir=None, albums=False, timeline=False,
            *args, **kwargs):
    json = Json(Config(config_file), output_dir=output_dir)
    if not albums and not timeline:
        albums, timeline = True, True
    json.generate(albums=albums, timeline=timeline)



META_EXAMPLE = """
# The start and end dates in the YYYY-MM-DD format.
start_date: 2016-02-01
end_date: 2016-02-02

# The title of your album
title: The title of your album

# In which order do you want to display photos in this album. Default is
# by file name or by date and both can be ascending or descending.
# name_asc (default), name_desc, date_asc, date_desc
order_by: name_asc

# A list of subdirectories in the album directory that you want to exclude.
# To exclude the top directory just use: "- "
exclude:
    - rejected

# The file name of the cover photo for this album. The cover photo is the
# photo that is shown on the album overview page.
cover: IMG_000000.JPG

# You can divide your album in chapters. This can be done automatically or
# manually by providing the following settings.
chapters:
    - title: First chapter
      starts: IMG_000001.JPG
      subtitle: A small subtitle.
    - title: Second chapter.
      starts: IMG_000002.JPG
      # Add a daily subtitle automatically
      subtitle: daily
      # You can specify the anchor of this chapter. That way you can link to
      # it from your description
      anchor: second-chapter
      # etc.

# You can add your own parameters. They will be included in the json file
# for the album and you can use them on your site.
your_own_param: awesome

---
# Nice Title

In this part of the file you can use a markdown formatted description
of your project.

You can link to your chapters. E.g. [to the second one](#second-chapter). Or
you just link [to the firs chapter](#first-chapter).
"""

META_EXAMPLE_SHORT = """
title: %(title)s
start_date: %(start_date)s
# end_date:
# cover:
# chapters: daily
%(extra)s
"""

META_EXAMPLE_NO_ALBUM = """
album: false
"""



# coding: utf-8

import click
import datetime
import os
import peewee
import pickle
import string

from ..models import Album
from ..models import AlbumPicture
from ..models import Picture
from ..models import ResizedPicture
from ..utils import check_and_create_dir
from ..utils import fatal_error
from ..utils import slugify
from ..utils import to_utc_timestamp
from ..utils import write_json


class Json(object):

    copy_failed_for = []
    files_to_copy = []
    templates = []
    anchors = []

    def __init__(self, config, output_dir=None):
        self.config = config
        self.output_dir = output_dir if output_dir else config.json_data_dir

    def _anchor(self, text):

        def _a(anchor, counter):
            return '%s-%s' % (anchor, counter) if counter else anchor

        anchor = slugify(text)
        counter = 0
        while _a(anchor, counter) in self.anchors:
            counter += 1
        return _a(anchor, counter)

    def _generate_albums_json(self):
        abs_dir = os.path.join(self.output_dir, 'albums')
        if not check_and_create_dir(abs_dir, create=True):
            return None
        album_titles, album_slugs, album_list = [], [], []
        albums = Album.select().where(Album.hidden == False)\
            .order_by(Album.start_date.desc())
        with click.progressbar(albums, label="Albums exported: ",
                               length=albums.count(), show_pos=True) as bar:
            for album in bar:
                pics = None
                content = []
                meta = pickle.loads(album.meta.encode()) if album.meta else {}
                if album.from_dir and 'name_desc' == meta.get('order_by', ''):
                    pics = album.pictures.order_by(Picture.file_name.desc())
                elif album.from_dir and 'date_desc' == meta.get('order_by', ''):
                    pics = album.pictures.order_by(Picture.sort_date.desc())
                elif album.from_dir and 'date_asc' == meta.get('order_by', ''):
                    pics = album.pictures.order_by(Picture.sort_date.asc())
                elif album.from_dir:
                    pics = album.pictures.order_by(Picture.file_name.asc())
                else:
                    one_day = datetime.timedelta(1)
                    if album.end_date is None:
                        end_date = album.start_date + one_day
                    else:
                        end_date = album.end_date + one_day
                    pics = Picture.select().where(
                        Picture.sort_date > album.start_date,
                        Picture.sort_date < end_date
                    ).order_by(Picture.sort_date.asc())

                chapters = meta.get('chapters', None)
                daily_titles, daily_subs = False, False
                if isinstance(chapters, list):
                    daily_titles = False
                elif chapters == 'daily' or self.config.auto_day_paragraphs:
                    daily_titles = True

                cover = meta.get('cover', None)
                if cover:
                    try:
                        cover = Picture.get(Picture.file_name == cover)\
                            .build_dict(self.config.resized_photos_url)
                    except peewee.DoesNotExist:
                        cover = None

                def _comparable_day(date, alt=0):
                    if isinstance(date, datetime.date):
                        return int(datetime.datetime.strftime(date, "%Y%m%d"))
                    return alt

                if pics is not None:
                    previous_day = 0
                    for pi, pic in enumerate(pics):
                        new_chapter = None
                        day = _comparable_day(pic.sort_date, previous_day)
                        if isinstance(chapters, list):
                            for c in chapters:
                                if c['starts'] == pic.file_name:
                                    new_chapter = c
                            if new_chapter and new_chapter.get('title'):
                                anchor = new_chapter.get('anchor', None)
                                if anchor is None:
                                    anchor = self._anchor(new_chapter['title'])
                                content.append({
                                    'type': 'title',
                                    'text': new_chapter['title'],
                                    'anchor': anchor,
                                })
                            if new_chapter and new_chapter.get('subtitle'):
                                if new_chapter['subtitle'] == 'daily':
                                    daily_subs = True
                                    content.append({
                                        'type': 'subtitle',
                                        'date': to_utc_timestamp(pic.sort_date),
                                        'anchor': pic.sort_date.strftime(
                                            "day-%Y-%m-%d"
                                        ),
                                    })
                                else:
                                    daily_subs = False
                                    content.append({
                                        'type': 'subtitle',
                                        'text': new_chapter['subtitle'],
                                        'anchor': self._anchor(
                                            new_chapter['subtitle']
                                        ),
                                    })
                            elif new_chapter:
                                daily_subs = False

                        if (daily_titles or daily_subs) and not new_chapter \
                                and day > previous_day:
                            content.append({
                                'type': 'subtitle' if daily_subs else 'title',
                                'date': to_utc_timestamp(pic.sort_date),
                                'anchor': pic.sort_date.strftime(
                                    "day-%Y-%m-%d"
                                ),
                            })

                        pic_dict = pic.build_dict(
                            self.config.resized_photos_url
                        )

                        # If we do not have a cover, we take the first picture
                        if cover is None:
                            cover = pic_dict.copy()

                        for resize in pic.resized_pictures.where(
                            ResizedPicture.successfully_created
                        ).where(
                            ResizedPicture.size << self.config.sizes_in_json
                        ):
                            pic_dict[resize.size.name] = [
                                resize.width,
                                resize.height,
                            ]
                        content.append(pic_dict)
                        previous_day = day

                    json_name = '%s.json' % album.slug
                    album_titles.append(string.capwords(album.title))
                    album_slugs.append(album.slug)
                    album_dict = {
                        'title': string.capwords(album.title),
                        'start_date': to_utc_timestamp(album.start_date),
                        'end_date': to_utc_timestamp(album.end_date),
                        'json': os.path.join('albums', json_name),
                        'slug': album.slug,
                        'total': pics.count(),
                        'cover': cover,
                    }
                    album_list.append(album_dict.copy())
                    album_dict['content'] = content
                    if album.description:
                        album_dict['description'] = album.description
                    if album.meta:
                        album_dict['meta'] = pickle.loads(album.meta.encode())
                    write_json(os.path.join(abs_dir, json_name), album_dict)
        write_json(
            os.path.join(self.output_dir, 'albums.json'),
            {
                'titles': album_titles,
                'albums': album_list,
                'slugs': album_slugs,
            },
        )

    def _generate_timeline_page(self, year=None, include_album_photos=False):
        if include_album_photos:
            base_query = Picture.select()\
                .where(Picture.sort_date.is_null(False))
        else:
            base_query = Picture.select()\
                .join(AlbumPicture, peewee.JOIN.LEFT_OUTER)\
                .join(Album, peewee.JOIN.LEFT_OUTER)\
                .where(Album.id.is_null())\
                .where(Picture.sort_date.is_null(False))

        if year is not None:
            base_query = base_query.where(
                (Picture.sort_date >= datetime.datetime(year, 1, 1)) &
                (Picture.sort_date < datetime.datetime(year + 1, 1, 1))
            )

        base_query = base_query.order_by(Picture.sort_date.desc())

        total = base_query.count()
        if not total:
            return None

        # Create the destination directories
        dir_name = 'including' if include_album_photos else 'excluding'
        dest = os.path.join(self.output_dir, 'timeline', dir_name)
        if not check_and_create_dir(dest, create=True):
            return None
        if year is not None:
            dest = os.path.join(dest, "%s" % year)
            if not check_and_create_dir(dest, create=True):
                return None

        nr_of_pages = int(total / self.config.photos_per_json_file) + \
            (1 if total % self.config.photos_per_json_file else 0)

        def _get_timeline_pics(base_query, page=0):
            page += 1
            pics = base_query.paginate(
                page, self.config.photos_per_json_file
            )
            return pics, page

        if self.config.timeline_granularity == 'daily':
            l1, l2, f1, f2 = 'month', 'day', 'month-%Y-%m', 'day-%Y-%m-%d'
        else:
            l1, l2, f1, f2 = 'year', 'month', 'year-%Y', 'month-%Y-%m'

        pages = []
        pics, page = _get_timeline_pics(base_query)
        previous_date = None
        label = '{:>7}'.format("%s | " % year if year else " All | ")
        label += "including:" if include_album_photos else "excluding:"
        with click.progressbar(length=nr_of_pages, label=label,
                               show_pos=True) as bar:
            while pics.count():
                content = []
                for pic in pics:
                    date = pic.sort_date
                    timestamp = to_utc_timestamp(date)
                    if previous_date is None or not \
                            getattr(date, l1) == getattr(previous_date, l1):
                        content.append({
                            'type': "title-%s" % l1,
                            'date': timestamp,
                            'anchor': date.strftime(f1),
                        })
                    if previous_date is None or not \
                            getattr(date, l2) == getattr(previous_date, l2):
                        content.append({
                            'type': "subtitle-%s" % l2,
                            'date': timestamp,
                            'anchor': date.strftime(f2),
                        })

                    pic_dict = pic.build_dict(self.config.resized_photos_url)

                    for resize in pic.resized_pictures.where(
                        ResizedPicture.successfully_created
                    ).where(
                        ResizedPicture.size << self.config.sizes_in_json
                    ):
                        pic_dict[resize.size.name] = [
                            resize.width,
                            resize.height,
                        ]

                    content.append(pic_dict)
                    previous_date = date
                if not pages:
                    pages.append(to_utc_timestamp(pics[0].sort_date))
                pages.append(to_utc_timestamp(pics[pics.count() - 1].sort_date))
                write_json(
                    os.path.join(dest, '%s.json' % page),
                    {'content': content, 'page': page, },
                )
                bar.update(1)
                pics, page = _get_timeline_pics(base_query, page=page)
        return pages

    def _generate_timeline_json(self):
        timeline_dir = os.path.join(self.output_dir, 'timeline')
        if not check_and_create_dir(timeline_dir, create=True):
            return fatal_error('Could not create Timeline directory.')
        timeline = {
            'including': {'years': {}, },
            'excluding': {'years': {}, },
            'page_dir': 'timeline',
        }
        first_year = Picture.select()\
            .where(Picture.sort_date.is_null(False))\
            .order_by(Picture.sort_date.asc())\
            .limit(1)[0].sort_date.year
        last_year = Picture.select()\
            .where(Picture.sort_date.is_null(False))\
            .order_by(Picture.sort_date.desc())\
            .limit(1)[0].sort_date.year
        # Generate complete timeline:
        # 1. Including album photos
        timeline['including']['pages'] = self._generate_timeline_page(
            include_album_photos=True
        )
        # 2. Excluding album photos
        timeline['excluding']['pages'] = self._generate_timeline_page()
        # The same per year.
        for year in range(first_year, last_year + 1):
            timeline['including']['years'][year] = \
                self._generate_timeline_page(year=year,
                                             include_album_photos=True)
            timeline['excluding']['years'][year] = \
                self._generate_timeline_page(year=year)
        write_json(os.path.join(self.output_dir, 'timeline.json'), timeline)

    def generate(self, albums=True, timeline=True):
        # Generate JSON
        if albums:
            click.echo("Generating json for albums... ", nl=False)
            self._generate_albums_json()
            click.echo("Done.")
        if timeline:
            click.echo("Generating json for the timeline... ", nl=False)
            self._generate_timeline_json()
            click.echo("Done.")
        click.echo("Exported json files to: %s" % self.output_dir)
